///////////////////////////////////////////////////////////////////////////////
//
/// \file       filter_flags_encoder.c
/// \brief      Decodes a Filter Flags field
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "common.h"
#include "lzma_encoder.h"
#include "fastpos.h"


/// \brief      Calculates the size of the Filter Properties field
///
/// This currently can return only LZMAALPHA_OK or LZMAALPHA_HEADER_ERROR, but
/// with some new filters it may return also LZMAALPHA_PROG_ERROR.
static lzmaalpha_ret
get_properties_size(uint32_t *size, const lzmaalpha_options_filter *options)
{
	lzmaalpha_ret ret = LZMAALPHA_OK;

	switch (options->id) {
#ifdef HAVE_FILTER_COPY
	case LZMAALPHA_FILTER_COPY:
		*size = 0;
		break;
#endif

#ifdef HAVE_FILTER_SUBBLOCK
	case LZMAALPHA_FILTER_SUBBLOCK:
		*size = 0;
		break;
#endif

#ifdef HAVE_FILTER_SIMPLE
#	ifdef HAVE_FILTER_X86
	case LZMAALPHA_FILTER_X86:
#	endif
#	ifdef HAVE_FILTER_POWERPC
	case LZMAALPHA_FILTER_POWERPC:
#	endif
#	ifdef HAVE_FILTER_IA64
	case LZMAALPHA_FILTER_IA64:
#	endif
#	ifdef HAVE_FILTER_ARM
	case LZMAALPHA_FILTER_ARM:
#	endif
#	ifdef HAVE_FILTER_ARMTHUMB
	case LZMAALPHA_FILTER_ARMTHUMB:
#	endif
#	ifdef HAVE_FILTER_SPARC
	case LZMAALPHA_FILTER_SPARC:
#	endif
		if (options->options == NULL || ((const lzmaalpha_options_simple *)(
				options->options))->start_offset == 0)
			*size = 0;
		else
			*size = 4;
		break;
#endif

#ifdef HAVE_FILTER_DELTA
	case LZMAALPHA_FILTER_DELTA:
		*size = 1;
		break;
#endif

#ifdef HAVE_FILTER_LZMA
	case LZMAALPHA_FILTER_LZMA:
		*size = 2;
		break;
#endif

	default:
		// Unknown filter - if the Filter ID is a proper VLI,
		// return LZMAALPHA_HEADER_ERROR instead of LZMAALPHA_PROG_ERROR,
		// because it's possible that we just don't have support
		// compiled in for the requested filter.
		ret = options->id <= LZMAALPHA_VLI_VALUE_MAX
				? LZMAALPHA_HEADER_ERROR : LZMAALPHA_PROG_ERROR;
		break;
	}

	return ret;
}


extern LZMA_API lzmaalpha_ret
lzmaalpha_filter_flags_size(uint32_t *size, const lzmaalpha_options_filter *options)
{
	// Get size of Filter Properties.
	uint32_t prop_size;
	const lzmaalpha_ret ret = get_properties_size(&prop_size, options);
	if (ret != LZMAALPHA_OK)
		return ret;

	// Size of Filter ID field if it exists.
	size_t id_size;
	size_t prop_size_size;
	if (options->id < 0xE0
			&& (lzmaalpha_vli)(prop_size) == options->id / 0x20) {
		// ID and Size of Filter Properties fit into Misc.
		id_size = 0;
		prop_size_size = 0;

	} else {
		// At least Filter ID is stored using the External ID field.
		id_size = lzmaalpha_vli_size(options->id);
		if (id_size == 0)
			return LZMAALPHA_PROG_ERROR;

		if (prop_size <= 30) {
			// Size of Filter Properties fits into Misc still.
			prop_size_size = 0;
		} else {
			// The Size of Filter Properties field is used too.
			prop_size_size = lzmaalpha_vli_size(prop_size);
			if (prop_size_size == 0)
				return LZMAALPHA_PROG_ERROR;
		}
	}

	// 1 is for the Misc field.
	*size = 1 + id_size + prop_size_size + prop_size;

	return LZMAALPHA_OK;
}


#ifdef HAVE_FILTER_SIMPLE
/// Encodes Filter Properties of the so called simple filters
static lzmaalpha_ret
properties_simple(uint8_t *out, size_t *out_pos, size_t out_size,
		const lzmaalpha_options_simple *options)
{
	if (options == NULL || options->start_offset == 0)
		return LZMAALPHA_OK;

	if (out_size - *out_pos < 4)
		return LZMAALPHA_BUF_ERROR;

	for (size_t i = 0; i < 4; ++i)
		out[(*out_pos)++] = options->start_offset >> (i * 8);

	return LZMAALPHA_OK;
}
#endif


#ifdef HAVE_FILTER_DELTA
/// Encodes Filter Properties of the Delta filter
static lzmaalpha_ret
properties_delta(uint8_t *out, size_t *out_pos, size_t out_size,
		const lzmaalpha_options_delta *options)
{
	if (options == NULL)
		return LZMAALPHA_PROG_ERROR;

	// It's possible that newer liblzma versions will support larger
	// distance values.
	if (options->distance < LZMAALPHA_DELTA_DISTANCE_MIN
			|| options->distance > LZMAALPHA_DELTA_DISTANCE_MAX)
		return LZMAALPHA_HEADER_ERROR;

	if (out_size - *out_pos < 1)
		return LZMAALPHA_BUF_ERROR;

	out[*out_pos] = options->distance - LZMAALPHA_DELTA_DISTANCE_MIN;
	++*out_pos;

	return LZMAALPHA_OK;
}
#endif


#ifdef HAVE_FILTER_LZMA
/// Encodes LZMA Properties and Dictionary Flags (two bytes)
static lzmaalpha_ret
properties_lzma(uint8_t *out, size_t *out_pos, size_t out_size,
		const lzmaalpha_options_lzma *options)
{
	if (options == NULL)
		return LZMAALPHA_PROG_ERROR;

	if (out_size - *out_pos < 2)
		return LZMAALPHA_BUF_ERROR;

	// LZMA Properties
	if (lzmaalpha_lzma_encode_properties(options, out + *out_pos))
		return LZMAALPHA_HEADER_ERROR;

	++*out_pos;

	// Dictionary flags
	//
	// Dictionary size is encoded using similar encoding that is used
	// internally by LZMA.
	//
	// This won't work if dictionary size can be zero:
#	if LZMAALPHA_DICTIONARY_SIZE_MIN < 1
#		error LZMAALPHA_DICTIONARY_SIZE_MIN cannot be zero.
#	endif

	uint32_t d = options->dictionary_size;

	// Validate it:
	if (d < LZMAALPHA_DICTIONARY_SIZE_MIN || d > LZMAALPHA_DICTIONARY_SIZE_MAX)
		return LZMAALPHA_HEADER_ERROR;

	// Round up to to the next 2^n or 2^n + 2^(n - 1) depending on which
	// one is the next:
	--d;
	d |= d >> 2;
	d |= d >> 3;
	d |= d >> 4;
	d |= d >> 8;
	d |= d >> 16;
	++d;

	// Get the highest two bits using the proper encoding:
	out[*out_pos] = get_pos_slot(d) - 1;
	++*out_pos;

	return LZMAALPHA_OK;
}
#endif


extern LZMA_API lzmaalpha_ret
lzmaalpha_filter_flags_encode(uint8_t *out, size_t *out_pos, size_t out_size,
		const lzmaalpha_options_filter *options)
{
	// Minimum output is one byte (everything fits into Misc).
	// The caller should have checked that there is enough output space,
	// so we return LZMAALPHA_PROG_ERROR instead of LZMAALPHA_BUF_ERROR.
	if (*out_pos >= out_size)
		return LZMAALPHA_PROG_ERROR;

	// Get size of Filter Properties.
	uint32_t prop_size;
	lzmaalpha_ret ret = get_properties_size(&prop_size, options);
	if (ret != LZMAALPHA_OK)
		return ret;

	// Misc, External ID, and Size of Properties
	if (options->id < 0xE0
			&& (lzmaalpha_vli)(prop_size) == options->id / 0x20) {
		// ID and Size of Filter Properties fit into Misc.
		out[*out_pos] = options->id;
		++*out_pos;

	} else if (prop_size <= 30) {
		// Size of Filter Properties fits into Misc.
		out[*out_pos] = prop_size + 0xE0;
		++*out_pos;

		// External ID is used to encode the Filter ID. If encoding
		// the VLI fails, it's because the caller has given as too
		// little output space, which it should have checked already.
		// So return LZMAALPHA_PROG_ERROR, not LZMAALPHA_BUF_ERROR.
		size_t dummy = 0;
		if (lzmaalpha_vli_encode(options->id, &dummy, 1,
				out, out_pos, out_size) != LZMAALPHA_STREAM_END)
			return LZMAALPHA_PROG_ERROR;

	} else {
		// Nothing fits into Misc.
		out[*out_pos] = 0xFF;
		++*out_pos;

		// External ID is used to encode the Filter ID.
		size_t dummy = 0;
		if (lzmaalpha_vli_encode(options->id, &dummy, 1,
				out, out_pos, out_size) != LZMAALPHA_STREAM_END)
			return LZMAALPHA_PROG_ERROR;

		// External Size of Filter Properties
		dummy = 0;
		if (lzmaalpha_vli_encode(prop_size, &dummy, 1,
				out, out_pos, out_size) != LZMAALPHA_STREAM_END)
			return LZMAALPHA_PROG_ERROR;
	}

	// Filter Properties
	switch (options->id) {
#ifdef HAVE_FILTER_COPY
	case LZMAALPHA_FILTER_COPY:
		assert(prop_size == 0);
		ret = options->options == NULL ? LZMAALPHA_OK : LZMAALPHA_HEADER_ERROR;
		break;
#endif

#ifdef HAVE_FILTER_SUBBLOCK
	case LZMAALPHA_FILTER_SUBBLOCK:
		assert(prop_size == 0);
		ret = LZMAALPHA_OK;
		break;
#endif

#ifdef HAVE_FILTER_SIMPLE
#	ifdef HAVE_FILTER_X86
	case LZMAALPHA_FILTER_X86:
#	endif
#	ifdef HAVE_FILTER_POWERPC
	case LZMAALPHA_FILTER_POWERPC:
#	endif
#	ifdef HAVE_FILTER_IA64
	case LZMAALPHA_FILTER_IA64:
#	endif
#	ifdef HAVE_FILTER_ARM
	case LZMAALPHA_FILTER_ARM:
#	endif
#	ifdef HAVE_FILTER_ARMTHUMB
	case LZMAALPHA_FILTER_ARMTHUMB:
#	endif
#	ifdef HAVE_FILTER_SPARC
	case LZMAALPHA_FILTER_SPARC:
#	endif
		ret = properties_simple(out, out_pos, out_size,
				options->options);
		break;
#endif

#ifdef HAVE_FILTER_DELTA
	case LZMAALPHA_FILTER_DELTA:
		ret = properties_delta(out, out_pos, out_size,
				options->options);
		break;
#endif

#ifdef HAVE_FILTER_LZMA
	case LZMAALPHA_FILTER_LZMA:
		ret = properties_lzma(out, out_pos, out_size,
				options->options);
		break;
#endif

	default:
		assert(0);
		ret = LZMAALPHA_PROG_ERROR;
		break;
	}

	return ret;
}
